import types
from trame import get_app_instance, trigger_key


def str_key_prefix(txt):
    if txt.startswith("{") or txt.startswith("`"):
        return ":"

    return ""


def py2js_key(key):
    return key.replace("_", "-")


class AbstractElement:
    _next_id = 1

    def __init__(self, name, __content=None, children=None, **kwargs):
        AbstractElement._next_id += 1
        self._id = AbstractElement._next_id
        self._elem_name = name
        self._allowed_keys = set()
        self._attr_names = []
        self._event_names = []

        self._attributes = {}
        self._txt = None
        self._py_attr = kwargs
        self._children = []

        if __content:
            if isinstance(__content, str):
                self._txt = __content
            elif isinstance(__content, list):
                self._children.extend(__content)
            else:
                self._children.append(__content)

        # Add standard Vue attribute handling
        if children:
            self._children.extend(children)

        # Add standard Vue attr/event handling
        self._attr_names += [
            "id",
            ("classes", "class"),
            "style",
            "v_model",
            "v_if",
            "v_show",
        ]
        self._event_names += ["click", "mousedown", "mouseup"]

    def _attr_str(self):
        return " ".join(self._attributes.values())

    def _update_allowed_keys(self):
        if hasattr(self, "_attr_names") and hasattr(self, "_event_names"):
            for items in [self._attr_names, self._event_names]:
                for item in items:
                    if isinstance(item, str):
                        self._allowed_keys.add(item)
                    else:
                        self._allowed_keys.add(item[0])

    # -------------------------------------------------------------------------
    # Buildin API
    # -------------------------------------------------------------------------

    def __getitem__(self, name):
        return self._py_attr[name]

    def __setitem__(self, name, value):
        if name in self._allowed_keys:
            self._py_attr[name] = value
        else:
            print(f"Attribute {name} is not defined for {self._elem_name}")

    def __getattr__(self, name):
        if name[0] == "_":
            raise AttributeError()

        return self._py_attr[name]

    def __setattr__(self, name, value):
        if name[0] == "_":
            self.__dict__[name] = value
        elif name == "content":
            self._txt = value
        elif name in self._allowed_keys:
            self._py_attr[name] = value
        else:
            self.__dict__[name] = value

        if name in ["_attr_names", "_event_names"]:
            self._update_allowed_keys()

    # -------------------------------------------------------------------------
    # helpers
    # -------------------------------------------------------------------------

    def ttsSensitive(self):
        self._attributes["__tts"] = f':key="`w{self._id}-${{tts}}`"'
        return self

    def attrs(self, *names):
        _app = get_app_instance()
        for _name in names:
            js_key = None
            name = _name
            if isinstance(_name, tuple):
                name = _name[0]
                js_key = _name[1]

            if name in self._py_attr:
                if js_key is None:
                    js_key = py2js_key(name)
                value = self._py_attr[name]

                if value is None:
                    continue

                if isinstance(value, (tuple, list)):
                    if len(value) > 1 and value[0] not in _app.state:
                        _app.state[value[0]] = value[1]

                    if js_key.startswith("v-"):
                        self._attributes[name] = f'{js_key}="{value[0]}"'
                    elif js_key.startswith(":"):
                        self._attributes[name] = f'{js_key}="{value[0]}"'
                    else:
                        self._attributes[name] = f':{js_key}="{value[0]}"'
                elif isinstance(value, bool):
                    if value:
                        self._attributes[name] = js_key
                elif isinstance(value, (str, int, float)):
                    self._attributes[name] = f'{js_key}="{value}"'
                else:
                    print(
                        f"Error: Don't know how to handle attribue name '{name}' with value '{value}' in {self.__class__}::{self._elem_name}"
                    )

        return self

    def events(self, *names):
        _app = get_app_instance()
        for _name in names:
            js_key = None
            name = _name
            if isinstance(_name, tuple):
                name = _name[0]
                js_key = _name[1]
            if name in self._py_attr:
                if js_key is None:
                    js_key = py2js_key(name)
                js_key = f"@{js_key}"
                value = self._py_attr[name]

                if value is None:
                    continue

                if isinstance(value, str):
                    self._attributes[name] = f'{js_key}="{value}"'
                elif isinstance(value, types.FunctionType):
                    trigger_name = trigger_key(value)
                    self._attributes[name] = f"{js_key}=\"trigger('{trigger_name}')\""
                elif isinstance(value, tuple):
                    trigger_name = value[0]
                    if isinstance(trigger_name, types.FunctionType):
                        trigger_name = trigger_key(trigger_name)
                    if len(value) == 1:
                        self._attributes[
                            name
                        ] = f"{js_key}=\"trigger('{trigger_name}')\""
                    if len(value) == 2:
                        self._attributes[
                            name
                        ] = f"{js_key}=\"trigger('{trigger_name}', {value[1]})\""
                    if len(value) == 3:
                        self._attributes[
                            name
                        ] = f"{js_key}=\"trigger('{trigger_name}', {value[1]}, {value[2]})\""
                else:
                    print(
                        f"Error: Don't know how to handle event name '{name}' with value '{value}' in {self.__class__}::{self._elem_name}"
                    )
        return self

    @property
    def content(self):
        return self._txt

    @property
    def children(self):
        return self._children

    @property
    def html(self):
        # Build attributes
        self.attrs(*self._attr_names)
        self.events(*self._event_names)

        # Compute HTML str
        if len(self._children):
            out_buffer = []
            out_buffer.append(f"<{self._elem_name} {self._attr_str()}>")
            for child in self._children:
                if isinstance(child, str):
                    out_buffer.append(child)
                else:
                    out_buffer.append(child.html)
            out_buffer.append(f"</{self._elem_name}>")
            return "\n".join(out_buffer)
        elif self._txt:
            return (
                f"<{self._elem_name} {self._attr_str()}>{self._txt}</{self._elem_name}>"
            )
        else:
            return f"<{self._elem_name} {self._attr_str()} />"


class Element(AbstractElement):
    def __init__(self, __elem_name, __content=None, **kwargs):
        super().__init__(__elem_name, __content, **kwargs)


class Div(AbstractElement):
    def __init__(self, __content=None, **kwargs):
        super().__init__("div", __content, **kwargs)


class Span(AbstractElement):
    def __init__(self, __content=None, **kwargs):
        super().__init__("span", __content, **kwargs)



from trame import get_app_instance
from trame.html import AbstractElement

# Make sure used module is available
_app = get_app_instance()
if "vuetify" not in _app.vue_use:
    _app.vue_use += ["vuetify"]

# v-alert
class VAlert(AbstractElement):
    def __init__(self, **kwargs):
        super().__init__("v-alert", **kwargs)
        self._attr_names += [
            "border",
            "color",
            "colored_border",
            "dark",
            "dense",
            "dismissible",
            "elevation",
            "height",
            "icon",
            "light",
            "max_height",
            "max_width",
            "min_height",
            "min_width",
            "mode",
            "origin",
            "outlined",
            "prominent",
            "rounded",
            "shaped",
            "tag",
            "text",
            "tile",
            "transition",
            "type",
            "value",
            "width",
        ]


# v-app
class VApp(AbstractElement):
    def __init__(self, **kwargs):
        super().__init__("v-app", **kwargs)
        self._attr_names += ["id"]


# v-app-bar
class VAppBar(AbstractElement):
    def __init__(self, **kwargs):
        super().__init__("v-app-bar", **kwargs)
        self._attr_names += [
            "absolute",
            "app",
            "bottom",
            "clipped_left",
            "clipped_right",
            "collapse",
            "collapse_on_scroll",
            "color",
            "dark",
            "dense",
            "elevate_on_scroll",
            "elevation",
            "extended",
            "extension_height",
            "fade_img_on_scroll",
            "fixed",
            "flat",
            "floating",
            "height",
            "hide_on_scroll",
            "inverted_scroll",
            "light",
            "max_height",
            "max_width",
            "min_height",
            "min_width",
            "outlined",
            "prominent",
            "rounded",
            "scroll_off_screen",
            "scroll_target",
            "scroll_threshold",
            "shaped",
            "short",
            "shrink_on_scroll",
            "src",
            "tag",
            "tile",
            "value",
            "width",
        ]


# v-app-bar-nav-icon
# v-app-bar-title
# v-autocomplete
# v-avatar
# v-badge
# v-banner
# v-bottom-navigation
# v-bottom-sheet
# v-breadcrumbs
# v-breadcrumbs-divider
# v-breadcrumbs-item
# v-btn
class VBtn(AbstractElement):
    def __init__(self, __content=None, **kwargs):
        super().__init__("v-btn", __content, **kwargs)
        self._attr_names += [
            "absolute",
            "active_class",
            "append",
            "block",
            "bottom",
            "color",
            "dark",
            "depressed",
            "disabled",
            "elevation",
            "exact",
            "exact_active_class",
            "exact_path",
            "fab",
            "fixed",
            "height",
            "href",
            "icon",
            "input_value",
            "large",
            "left",
            "light",
            "link",
            "loading",
            "max_height",
            "max_width",
            "min_height",
            "min_width",
            "outlined",
            "plain",
            "replace",
            "retain_focus_on_click",
            "right",
            "ripple",
            "rounded",
            "shaped",
            "small",
            "tag",
            "target",
            "text",
            "tile",
            "to",
            "top",
            "type",
            "value",
            "width",
            "x_large",
            "x_small",
        ]


# v-btn-toggle
class VBtnToggle(AbstractElement):
    def __init__(self, __content=None, **kwargs):
        super().__init__("v-btn-toggle", __content, **kwargs)
        self._attr_names += [
            "active_class",
            "background_color",
            "borderless",
            "color",
            "dark",
            "dense",
            "group",
            "light",
            "mandatory",
            "max",
            "multiple",
            "rounded",
            "shaped",
            "tag",
            "tile",
            "value",
        ]
        self._event_names += [
            "change",
        ]


# v-calendar
# v-calendar-daily
# v-calendar-monthly
# v-calendar-weekly
# v-card
# v-card-actions
# v-card-subtitle
# v-card-text
# v-card-title
# v-carousel
# v-carousel-item
# v-carousel-reverse-transition
# v-carousel-transition
# v-checkbox
# v-chip
# v-chip-group
# v-chip-outside
# v-col
class VCol(AbstractElement):
    def __init__(self, **kwargs):
        super().__init__("v-col", **kwargs)
        self._attr_names += [
            "align_self",
            "cols",
            "lg",
            "md",
            "offset",
            "offset_lg",
            "offset_md",
            "offset_sm",
            "offset_xl",
            "order",
            "order_lg",
            "order_md",
            "order_sm",
            "order_xl",
            "sm",
            "tag",
            "xl",
        ]


# v-color-picker
# v-combobox
# v-container
class VContainer(AbstractElement):
    def __init__(self, **kwargs):
        super().__init__("v-container", **kwargs)
        self._attr_names += ["id", "tag", "fluid", "fill_height"]


# v-content
class VContent(AbstractElement):
    def __init__(self, **kwargs):
        super().__init__("v-content", **kwargs)
        self._attr_names += ["tag"]


# v-data-footer
# v-data-iterator
# v-data-table
# v-data-table-header
# v-date-picker
# v-dialog
# v-dialog-bottom-transition
# v-dialog-top-transition
# v-dialog-transition
# v-divider
class VDivider(AbstractElement):
    def __init__(self, **kwargs):
        super().__init__("v-divider", **kwargs)
        self._attr_names += ["dark", "inset", "light", "vertical"]


# v-edit-dialog
# v-expand-transition
# v-expand-x-transition
# v-expansion-panel
# v-expansion-panel-content
# v-expansion-panel-header
# v-expansion-panels
# v-fab-transition
# v-fade-transition
# v-file-input
# v-flex
# v-footer
# v-form
# v-hover
# v-icon
class VIcon(AbstractElement):
    def __init__(self, __content, **kwargs):
        super().__init__("v-icon", __content, **kwargs)
        self._attr_names += [
            "color",
            "dark",
            "dense",
            "disabled",
            "large",
            "left",
            "light",
            "right",
            "size",
            "small",
            "tag",
            "x_large",
            "x_small",
        ]


# v-img
# v-input
# v-intersect
# v-item
# v-item-group
# v-layout
# v-lazy
# v-list
# v-list-group
# v-list-item
# v-list-item-action
# v-list-item-action-text
# v-list-item-avatar
# v-list-item-content
# v-list-item-group
# v-list-item-icon
# v-list-item-subtitle
# v-list-item-title
# v-main
class VMain(AbstractElement):
    def __init__(self, **kwargs):
        super().__init__("v-main", **kwargs)
        self._attr_names += ["tag"]


# v-menu
# v-menu-transition
# v-mutate
# v-navigation-drawer
# v-overflow-btn
# v-overlay
# v-pagination
# v-parallax
# v-progress-circular
# v-progress-linear
class VProgressLinear(AbstractElement):
    def __init__(self, **kwargs):
        super().__init__("v-progress-linear", **kwargs)
        self._attr_names += [
            "absolute",
            "active",
            "background_color",
            "background_opacity",
            "bottom",
            "buffer_value",
            "color",
            "dark",
            "fixed",
            "height",
            "indeterminate",
            "light",
            "query",
            "reverse",
            "rounded",
            "stream",
            "striped",
            "top",
            "value",
        ]
        self._event_names += ["change"]


# v-radio
# v-radio-group
# v-range-slider
# v-rating
# v-resize
# v-responsive
# v-ripple
# v-row
class VRow(AbstractElement):
    def __init__(self, **kwargs):
        super().__init__("v-row", **kwargs)
        self._attr_names += [
            "align",
            "align_content",
            "align_content_lg",
            "align_content_md",
            "align_content_sm",
            "align_content_xl",
            "align_lg",
            "align_md",
            "align_sm",
            "align_xl",
            "dense",
            "justify",
            "justify_lg",
            "justify_md",
            "justify_sm",
            "justify_xl",
            "no_gutter",
            "tag",
        ]


# v-scale-transition
# v-scroll
# v-scroll-x-reverse-transition
# v-scroll-x-transition
# v-scroll-y-reverse-transition
# v-scroll-y-transition
# v-select
class VSelect(AbstractElement):
    def __init__(self, **kwargs):
        super().__init__("v-select", **kwargs)
        self.ttsSensitive()
        self._attr_names += [
            "append_icon",
            "append_outer_icon",
            "attach",
            "autofocus",
            "background_color",
            "cache_items",
            "chips",
            "clear_icon",
            "clearable",
            "color",
            "counter",
            "dark",
            "deletable_chips",
            "dense",
            "disable_lookup",
            "disabled",
            "eager",
            "error",
            "error_count",
            "error_messages",
            "filled",
            "flat",
            "full_width",
            "height",
            "hide_details",
            "hide_selected",
            "hint",
            "id",
            "item_color",
            "item_disabled",
            "item_text",
            "item_value",
            "items",
            "label",
            "light",
            "loader_height",
            "loading",
            "menu_props",
            "messages",
            "multiple",
            "open_on_clear",
            "outlined",
            "persitent_hint",
            "persitent_placeholder",
            "placeholder",
            "prefix",
            "prepend_icon",
            "prepend_inner_icon",
            "readonly",
            "return_object",
            "reverse",
            "rounded",
            "rules",
            "shaped",
            "single_line",
            "small_chips",
            "solo",
            "solo_inverted",
            "success",
            "success_messages",
            "suffix",
            "type",
            "validate_on_blur",
            "value",
        ]
        self._event_names += [
            "blur",
            "change",
            "click",
            "click_append",
            "click_append_outer",
            "click_clear",
            "click_prepend",
            "click_prepend_inner",
            "focus",
            "input",
        ]


# v-sheet
# v-simple-checkbox
# v-simple-table
# v-skeleton-loader
# v-slide-group
# v-slide-item
# v-slide-x-reverse-transition
# v-slide-x-transition
# v-slide-y-reverse-transition
# v-slide-y-transition
# v-slider
class VSlider(AbstractElement):
    def __init__(self, **kwargs):
        super().__init__("v-slider", **kwargs)
        self._attr_names += [
            "append_icon",
            "background_color",
            "color",
            "dark",
            "dense",
            "disabled",
            "error",
            "error_count",
            "error_messages",
            "height",
            "hide_details",
            "hint",
            "id",
            "inverse_label",
            "label",
            "light",
            "loader_height",
            "loading",
            "max",
            "messages",
            "min",
            "persistent_hint",
            "prepend_icon",
            "readonly",
            "rules",
            "step",
            "success",
            "success_messages",
            "thumb_color",
            "thumb_label",
            "thumb_size",
            "tick_labels",
            "tick_size",
            "ticks",
            "track_color",
            "track_fill_color",
            "validate_on_blur",
            "value",
            "vertical",
        ]
        self._event_names += [
            "change",
            "click",
            ("click_append", "click:append"),
            ("click_prepend", "click:prepend"),
            "end",
            "input",
            "start",
            "mousedown",
            "mouseup",
        ]


# v-snackbar
# v-spacer
class VSpacer(AbstractElement):
    def __init__(self, **kwargs):
        super().__init__("v-spacer", **kwargs)


# v-sparkline
# v-speed-dial
# v-stepper
# v-stepper-content
# v-stepper-header
# v-stepper-items
# v-stepper-step
# v-subheader
# v-switch
class VSwitch(AbstractElement):
    def __init__(self, **kwargs):
        super().__init__("v-switch", **kwargs)
        self.ttsSensitive()
        self._attr_names += [
            "append_icon",
            "background_color",
            "color",
            "dark",
            "dense",
            "disabled",
            "error",
            "error_count",
            "error_messages",
            "false_value",
            "flat",
            "hide_details",
            "hint",
            "id",
            "input_value",
            "inset",
            "label",
            "light",
            "loading",
            "messages",
            "multiple",
            "persistent_hint",
            "prepend_icon",
            "readonly",
            "ripple",
            "rules",
            "success",
            "success_messages",
            "true_value",
            "validate_on_blur",
            "value",
        ]
        self._event_names += [
            "change",
            "click",
            ("click_append", "click:append"),
            ("click_prepend", "click:append"),
            "mousedown",
            "mouseup",
            ("update_error", "update:error"),
        ]


# v-system-bar
# v-tab
# v-tab-item
# v-tab-reverse-transition
# v-tab-transition
# v-tabs
# v-tabs-items
# v-tabs-slider
# v-text-field
# v-textarea
# v-theme-provider
# v-time-picker
# v-timeline
# v-timeline-item
# v-toolbar
# v-toolbar-items
# v-toolbar-title
# v-tooltip
# v-touch
# v-tree-view
# v-virtual-scroll
# v-window
# v-window-item



from trame import get_app_instance
from trame.html import AbstractElement

MODULE = None


def use_module(m):
    global MODULE
    MODULE = m
    _app = get_app_instance()
    _app.enableModule(m)


class VtkAlgorithm(AbstractElement):
    def __init__(self, __content=None, **kwargs):
        super().__init__("vtk-algorithm", __content, **kwargs)
        self._attr_names += ["port", "vtkClass", "state"]


class VtkCellData(AbstractElement):
    def __init__(self, __content=None, **kwargs):
        super().__init__("vtk-cell-data", __content, **kwargs)


class VtkDataArray(AbstractElement):
    def __init__(self, **kwargs):
        super().__init__("vtk-data-array", **kwargs)
        self._attr_names += [
            "name",
            "registration",
            "type",
            "values",
            "numberOfComponents",
        ]


class VtkFieldData(AbstractElement):
    def __init__(self, __content=None, **kwargs):
        super().__init__("vtk-field-data", __content, **kwargs)


class VtkGeometryRepresentation(AbstractElement):
    def __init__(self, __content=None, **kwargs):
        super().__init__("vtk-geometry-representation", __content, **kwargs)
        self._attr_names += [
            "id",
            "colorMapPreset",
            "colorDataRange",
            "actor",
            "mapper",
            "property",
        ]


class VtkGlyphRepresentation(AbstractElement):
    def __init__(self, __content=None, **kwargs):
        super().__init__("vtk-glyph-representation", __content, **kwargs)
        self._attr_names += [
            "colorMapPreset",
            "colorDataRange",
            "actor",
            "mapper",
            "property",
        ]


class VtkMesh(AbstractElement):
    def __init__(self, __content=None, **kwargs):
        super().__init__("vtk-mesh", __content, **kwargs)
        self._attr_names += ["port", "state"]


class VtkPointData(AbstractElement):
    def __init__(self, __content=None, **kwargs):
        super().__init__("vtk-point-data", __content, **kwargs)


class VtkPolyData(AbstractElement):
    def __init__(self, name, __content=None, dataset=None, **kwargs):
        super().__init__("vtk-polydata", __content, **kwargs)
        self.__name = name
        self.__dataset = dataset
        self._attr_names += [
            "port",
            "verts",
            "lines",
            "polys",
            "strips",
            "connectivity",
        ]
        if dataset:
            self._attributes["bind"] = f'v-bind="{name}.mesh"'
            self.update()

    def set_dataset(self, dataset):
        self.__dataset = dataset
        self.update()

    def update(self):
        if self.__dataset:
            _app = get_app_instance()
            _app.set(self.__name, MODULE.mesh(self.__dataset))


class VtkReader(AbstractElement):
    def __init__(self, **kwargs):
        super().__init__("vtk-reader", **kwargs)
        self._attr_names += [
            "parseAsArrayBuffer",
            "parseAsText",
            "port",
            "renderOnUpdate",
            "resetCameraOnUpdate",
            "url",
            "vtkClass",
        ]


class VtkRemoteLocalView(AbstractElement):
    def __init__(self, view, **kwargs):
        super().__init__("vtk-remote-local-view", **kwargs)
        __ns = kwargs["namespace"]

        # !!! HACK !!!
        # Allow user to configure view mode by providing (, local/remote)
        __rmode = "remote"
        __mode_arg = kwargs.get("mode", False)
        if __mode_arg and isinstance(__mode_arg, (tuple, list)) and len(__mode_arg) > 1:
            __rmode = __mode_arg[1]
            self._attributes["mode"] = f':mode="{__mode_arg[0]}"'
        elif __mode_arg and isinstance(__mode_arg, (tuple, list)):
            self._attributes["mode"] = f':mode="{__mode_arg[0]}"'
        elif __mode_arg and isinstance(__mode_arg, str):
            self._attributes["mode"] = f'mode="{__mode_arg}"'
        # !!! HACK !!!

        self.__scene_id = f"{__ns}Scene"
        self.__view_id = MODULE.id(view)
        self.__view = view
        self.__wrapped_view = MODULE.view(view, __ns, mode=__rmode)

        self._attributes["wsClient"] = ':wsClient="wsClient"'
        self._attributes["ref"] = f'ref="{kwargs.get("ref", __ns)}"'
        self._attributes["view_id"] = f'id="{self.__view_id}"'
        self._attributes["view_state"] = f':viewState="{self.__scene_id}"'

        self._attr_names += [
            # "mode", # <--- Managed by hand above
            "contextName",
            "interactiveRatio",
            "interactorEvents",
            "interactorSettings",
            "namespace",
        ]
        self._event_names += kwargs.get("interactorEvents", [])

    def update_geometry(self):
        _app = get_app_instance()
        _app.set(self.__scene_id, MODULE.scene(self.__view))

    def update_image(self):
        MODULE.push_image(self.__view)

    @property
    def view(self):
        return self.__wrapped_view


class VtkRemoteView(AbstractElement):
    @staticmethod
    def push_image(view):
        MODULE.push_image(view)

    def __init__(self, view, ref="view", **kwargs):
        super().__init__("vtk-remote-view", **kwargs)
        self.__view = view
        self._attributes["wsClient"] = ':wsClient="wsClient"'
        self._attributes["ref"] = f'ref="{ref}"'
        self._attributes["view_id"] = f'id="{MODULE.id(view)}"'
        self._attr_names += [
            "enablePicking",
            "interactiveQuality",
            "interactiveRatio",
            "interactorEvents",
        ]
        self._event_names += kwargs.get("interactorEvents", [])

    def update(self):
        MODULE.push_image(self.__view)


class VtkShareDataset(AbstractElement):
    def __init__(self, __content=None, **kwargs):
        super().__init__("vtk-share-dataset", __content, **kwargs)
        self._attr_names += ["port", "name"]


class VtkSyncView(AbstractElement):
    def __init__(self, view, ref="view", **kwargs):
        super().__init__("vtk-sync-view", **kwargs)
        self.__scene_id = f"scene_{self._id}"
        self.__view = view
        self._attributes["ref"] = f'ref="{ref}"'
        self._attributes["wsClient"] = ':wsClient="wsClient"'
        self._attributes["view_state"] = f':viewState="{self.__scene_id}"'
        self._attr_names += ["interactorEvents", "interactorSettings", "contextName"]
        self._event_names += kwargs.get("interactorEvents", [])
        self.update()

    def update(self):
        _app = get_app_instance()
        _app.set(self.__scene_id, MODULE.scene(self.__view))


class VtkView(AbstractElement):
    def __init__(self, __content=None, ref="view", **kwargs):
        super().__init__("vtk-view", __content, **kwargs)
        self._attributes["ref"] = f'ref="{ref}"'
        self._attr_names += [
            "background",
            "cubeAxesStyle",
            "interactorSettings",
            "pickingModes",
            "showCubeAxes",
        ]



from trame.html import Span, vuetify


class SinglePage:
    def __init__(self, name):
        self.name = name
        self._app = vuetify.VApp(id="app")
        self.toolbar = vuetify.VAppBar(app=True)
        self.logo = vuetify.VIcon("mdi-menu", classes="mr-4")
        self.title = Span("Trame App", classes="title")
        self.content = vuetify.VMain()
        self.toolbar.children += [self.logo, self.title]
        self._app.children += [self.toolbar, self.content]

    @property
    def html(self):
        return self._app.html


class FullScreenPage:
    def __init__(self, name):
        self.name = name
        self._app = vuetify.VApp(id="app")
        self.children = self._app.children

    @property
    def html(self):
        return self._app.html



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import re

from setuptools import setup, find_packages

# perform the install
setup(
    name="trame",
    description="Framework for building interactive Web applications quickly in plain Python",
    long_description="""
    Trame aims to be a framework for building interactive applications using a web front-end in plain Python. Such applications can be used locally as any desktop application but also be deployed in the cloud or on premise to access big and/or sensitive data.
    Trame comes with lots of capabilities built-in by leveraging existing libraries or tools such as Vuetify, Altair, Vega, Deck, VTK, ParaView and more. Trame lets you create interactive data processing applications with rich visualizations without switching languages or technologies.
    Trame lets you define your application in a very compact and intuitive way while letting you take over if you know better. Several available layouts let you build your application in no time.
    Trame is the perfect companion to VTK or ParaView by offering various rendering and processing configurations. Trame lets you choose between server-side and client-side rendering along with hybrid approaches.
    """,
    author="Kitware, Inc.",
    author_email="kitware@kitware.com",
    url=f"https://github.com/kitware/trame",
    license="BSD-3-Clause",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: JavaScript",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords="Python Interactive Web Application Framework",
    packages=find_packages(".", exclude=("tests.*", "tests")),
    package_dir={"": "."},
    package_data={},
    install_requires=["pywebvue"],
)



from trame import get_app_instance
from trame.html import AbstractElement

from pywebvue.modules import Markdown

# Make sure used module is available
_app = get_app_instance()
_app.enableModule(Markdown)


class Markdown(AbstractElement):
    def __init__(self, **kwargs):
        super().__init__("markdown-it-vue", **kwargs)
        self._attr_names += [("v_model", ":content")]



import os
import inspect

from pywebvue import App

NEXT_TRIGGER_ID = 0
TRIGGER_MAP = {}
NEXT_APP_ID = 0
APPS = {}
APP_STACK = []
APP = None


def base_directory():
    frame = inspect.stack()[2]
    module = inspect.getmodule(frame[0])
    if module is None:
        return os.getcwd()
    return os.path.abspath(os.path.dirname(module.__file__))


# -----------------------------------------------------------------------------
# App management
# -----------------------------------------------------------------------------


def get_app_instance():
    """
    Return the current active PyWebVue App instance and
    create one if none was available yet.

    (This method is meant for advanced users and should not be needed for most)
    """
    global APP
    if APP:
        return APP

    create_app("Trame Created Application")
    return APP


# -----------------------------------------------------------------------------


def activate_app(app_id):
    """
    When multiple application instances are use this method allow to toggle
    which app should be current based on the app_id.

    (This method is meant for advanced users and should not be needed for most)
    """
    global APP_STACK, APPS, APP
    if app_id in APPS:
        APP_STACK.append(app_id)
        APP = APPS[app_id]
        return True

    return False


# -----------------------------------------------------------------------------


def deactivate_app():
    """
    When multiple application instances are use this method allow to activate
    the previously activated app by desactivating the current one.

    (This method is meant for advanced users and should not be needed for most)
    """
    global APP_STACK, APPS, APP
    if len(APP_STACK):
        app_ip = APP_STACK.pop()
        APP = None
        if len(APP_STACK):
            activate_app(APP_STACK[-1])
        return app_ip


# -----------------------------------------------------------------------------


def create_app(name):
    """
    This will create a PyWebVue application instance,
    activate it and return its app_id so you could
    desactivate and reactivate it later.

    (This method is meant for advanced users and should not be needed for most)
    """
    global NEXT_APP_ID, APPS
    NEXT_APP_ID += 1
    _app_id = f"trame_app_{NEXT_APP_ID}"
    APPS[_app_id] = App(name)
    activate_app(_app_id)
    return _app_id


# -----------------------------------------------------------------------------
# Server start
# -----------------------------------------------------------------------------


def start(layout=None, name=None, favicon=None, on_ready=None, port=None):
    """
    Start web server for serving your application

    Parameters
    ----------
    layout  : None or str or trame.layouts.*
        UI content that should be used for your application
    name    : None or str
        "Title" that you can see in your tab browser.
        This will be filled automatically if a trame.layouts.* layout was provided.
    favicon : None or str
        Relative path to a png image that should be used as favicon
    port    : None or Number
        Port on which the server should run on
    on_ready: None or function
        Function called once the server is ready
    """
    app = get_app_instance()
    if name:
        app.name = name

    if favicon:
        app.favicon = os.path.join(base_directory(), favicon)

    if layout:
        if isinstance(layout, str):
            app.layout = layout
        else:
            app.name = layout.name
            app.layout = layout.html
    else:
        tpl_path = os.path.join(base_directory(), "template.html")
        if os.path.exists(tpl_path):
            app.layout = tpl_path

    app.on_ready = on_ready
    app.run_server(port=port)


# -----------------------------------------------------------------------------
# App state management
# -----------------------------------------------------------------------------


def update_state(key, value=None):
    """
    Method updating current application state that is shared with the Web UI
    """
    _app = get_app_instance()
    _app.set(key, value)


# -----------------------------------------------------------------------------


def get_state(*names):
    """
    Return a list of the values of the given state keys

    Paramters
    ---------
    *names : *str
        List of name of state values to retreive

    Returns
    -------
    [any, ...]
        List of value matching the requested state property names
    """
    _app = get_app_instance()
    results = []
    for name in names:
        results.append(_app.get(name))

    return results


def update_layout(layout):
    """
    Dynamically update current application layout

    Parameters
    ----------
    layout  : str or trame.layouts.*
        UI content that should be used for your application
    """
    _app = get_app_instance()
    _app.layout = layout if isinstance(layout, str) else layout.html


# -----------------------------------------------------------------------------
# Generate trigger helper
# -----------------------------------------------------------------------------


def trigger_key(_fn):
    """
    Providing a function, a generated trigger name will be returned

    Parameters
    ----------
    _fn : function
        Function that we would like to be able to call from the client side.

    Returns
    -------
    str
        Unique name that can be used for a trigger to call that function
    """
    # Return precomputed key
    global TRIGGER_MAP, NEXT_TRIGGER_ID
    if _fn in TRIGGER_MAP:
        return TRIGGER_MAP[_fn]

    # Compute unique trigger key
    NEXT_TRIGGER_ID += 1
    key = f"trigger_{NEXT_TRIGGER_ID}"
    TRIGGER_MAP[_fn] = key

    # Register function trigger
    _app = get_app_instance()
    _app.trigger(key)(_fn)

    return key


# -----------------------------------------------------------------------------
# @decorators
# -----------------------------------------------------------------------------


def change(*_args, **_kwargs):
    """
    @change decorator allow to register a function in a way that it will be
    automatically called when the given list of state names get modified.

    The decorated function is passed the full state as *kwarg when possible.
    This means you should have a method profile similar to `fn(..., *kwargs)`

    Parameters
    ----------
    *_args : *str
        List of name that your function should listen to
    """
    _app = get_app_instance()
    return _app.change(*_args, **_kwargs)


# -----------------------------------------------------------------------------


def trigger(name):
    """
    @trigger decorator allow to register a function as a trigger with a given name

    Parameters
    ----------
    name : str
        Name of the trigger
    """
    _app = get_app_instance()
    return _app.trigger(name)
